"""Module for creating a BDD feature file."""

import json
import sys  # Import sys
from pathlib import Path  # Import Path

from .collecting_formatter import CollectedFeature


def load_feature_file(feature_path: str):
    """Load a BDD feature file."""
    if not Path(feature_path).exists():
        print("Unable to find feature file\\n")
        sys.exit(2)  # Use sys.exit

    with open(feature_path) as file:
        return CollectedFeature.from_json(json.load(file))


def load_all_feature_files_in_directory(featurespath: str):
    """Load all BDD feature files in directory."""
    files = Path(featurespath).glob("*.json")

    features = []
    for file_path_obj in files:
        file = str(file_path_obj)
        feature = load_feature_file(file)
        features.append(feature)

    return features


status_to_style_dict: dict[str, str] = {
    "not run": "notrun",
    # behave.model_core.Status
    "untested": "notrun",
    "skipped": "notrun",
    "passed": "passed",
    "failed": "failed",
    "undefined": "notimplemented",
    "executing": "notrun",
}


def status_to_style(status: str):
    """Get a status style based on status"""
    return status_to_style_dict.get(status, "failed")



"""Behave JSONFormatter."""

import json
from pathlib import Path  # Import Path

from .collecting_formatter import CollectingFormatter


class JSONFormatter(CollectingFormatter):
    """Main class used by Behave framework."""

    def __init__(self, stream_opener, config):
        super().__init__(stream_opener, config)

        self.outputDirectory = Path(config.base_dir) / config.userdata["behave.formatter.customjson.path"]

        if "behave.formatter.customjson.single_file" in config.userdata:
            self.singleFile = config.userdata["behave.formatter.customjson.single_file"] == "True"
        else:
            self.singleFile = False

        self.features = []

    def eof(self):
        """End of file reached."""
        super().eof()

        if self.singleFile:
            self.features.append(self.current_feature)
        else:
            file_name = Path(self.current_feature.file_name).stem
            file_path = self.outputDirectory / (file_name + ".json")

            with open(file_path, "w") as file:
                json.dump(self.current_feature, file, default=lambda o: o.__dict__, indent=2)

    def close(self):
        """Dump content to file on close."""
        super().close()

        if self.singleFile:
            file_path = self.outputDirectory / "behave_test_results.json"
            with open(file_path, "w") as file:
                json.dump(self.features, file, default=lambda o: o.__dict__, indent=2)



"""Create Sphinx compatible feature page."""

import pathlib
import sys
from pathlib import Path  # Import Path

from jinja2 import Environment
from jinja2 import FileSystemLoader

from .collecting_formatter import CollectedStep
from .feature_file import load_feature_file
from .feature_file import status_to_style


current_directory = pathlib.Path(__file__).parent.resolve()
env = Environment(autoescape=True, loader=FileSystemLoader(current_directory, followlinks=True))
template = env.get_template("feature.jinja2")


def screenshots_from_step(step: CollectedStep):
    """Step screenshot."""
    screenshots = []
    for line in step.text:
        line_text: str = line.strip()
        line_split = line_text.split("'")
        if len(line_split) == 3 and line_split[0] == "Save screenshot ":
            screenshots.append(line_split[1])
    return screenshots


def main(args=None):
    """Entry point when running as script."""
    if args is None:
        args = sys.argv

    if len(args) != 2:
        print(f"Invalid command format, format is:\n {args[0]} <json feature result>\n")
        sys.exit(1)

    feature = load_feature_file(args[1])

    context = {
        "feature": feature,
        "status_to_style": status_to_style,
        "screenshots_from_step": screenshots_from_step,
    }

    file_path = Path(args[1])
    file_name = file_path.stem + ".rst"

    with open(file_name, "w") as file:
        file.write(template.render(**context))


def process_files_in_current_directory():
    """Process files."""
    files = Path().glob("*.json")

    for file_path_obj in files:
        file = str(file_path_obj)
        print(f"Converting {file}")
        feature = load_feature_file(file)

        context = {
            "feature": feature,
            "status_to_style": status_to_style,
            "screenshots_from_step": screenshots_from_step,
        }

        file_stem = Path(file).stem
        file_name = file_stem + ".rst"

        with open(file_name, "w") as f:
            f.write(template.render(**context))


if __name__ == "__main__":
    """If invoked directly as script, relay to main() method."""
    main()



"""Parse and format user needs."""

import csv


EXPECTED_NUMBER_OF_USERNEED_COLUMNS = 2


def load_userneeds_from_file(filename):
    """Load and parse user needs."""
    userneeds = []
    with open(filename) as tsv:
        next(tsv)  # Skip the title row
        for line in csv.reader(tsv, dialect="excel-tab"):
            if len(line) != EXPECTED_NUMBER_OF_USERNEED_COLUMNS:
                continue
            userneeds.append({"id": line[0], "description": line[1]})
    with open(filename) as csv_file:
        next(csv_file)  # Skip the title row
        for line in csv.reader(csv_file, dialect="excel"):
            if len(line) != EXPECTED_NUMBER_OF_USERNEED_COLUMNS:
                continue
            userneeds.append({"id": line[0], "description": line[1]})
    return userneeds



"""Implement Formatter functionality for Behave."""

import time
from itertools import chain

from behave.formatter.base import Formatter
from behave.model import Feature
from behave.model import Scenario
from behave.model import Step


class CollectedStep:
    """A BDD Step."""

    def __init__(
        self,
        name: str,
        step_type: str,
        text: list[str] | None = None,
        error_message: list[str] | None = None,
        status: str = "not run",
    ):
        self.name = name
        self.step_type = step_type
        self.text: list[str] = text if text else []
        self.error_message = error_message
        self.status = status

    @classmethod
    def from_json(cls, data):
        return cls(**data)


class CollectedScenario:
    """A BDD scenario."""

    def __init__(
        self,
        name: str,
        tags: list[str],
        status: str = "not run",
        steps: list[CollectedStep] | None = None,
    ):
        self.name = name
        self.tags: list[str] = tags
        self.status = status
        self.steps: list[CollectedStep] = [] if (steps is None) else steps

    @classmethod
    def from_json(cls, data):
        steps = list(map(CollectedStep.from_json, data["steps"]))
        data.pop("steps", None)
        return cls(steps=steps, **data)


class CollectedRule:
    """A BDD rule."""

    def __init__(
        self,
        name: str,
        description: list[str],
        tags: list[str],
        scenarios: list[CollectedScenario] | None = None,
    ):
        self.name = name
        self.description: list[str] = description
        self.tags: list[str] = tags
        self.scenarios: list[CollectedScenario] = scenarios if scenarios else []

    @classmethod
    def from_json(cls, data):
        scenarios = list(map(CollectedScenario.from_json, data["scenarios"]))
        data.pop("scenarios", None)
        return cls(scenarios=scenarios, **data)


class CollectedFeature:
    """A BDD Feature."""

    def __init__(
        self,
        file_name: str,
        name: str,
        description: list[str],
        tags: list[str],
        start_time: float | None = None,
        run_time: float = 0,
        scenarios: list[CollectedScenario] | None = None,
        rules: list[CollectedRule] | None = None,
    ):
        self.file_name = file_name
        self.name = name
        self.description: list[str] = description
        self.tags: list[str] = tags
        self.start_time = start_time if start_time else time.time()
        self.run_time = run_time
        self.scenarios: list[CollectedScenario] = [] if (scenarios is None) else scenarios
        self.rules: list[CollectedRule] = [] if (rules is None) else rules

    def finished(self):
        self.run_time = time.time() - self.start_time

    @classmethod
    def from_json(cls, data):
        scenarios = list(map(CollectedScenario.from_json, data["scenarios"]))
        data.pop("scenarios", None)
        rules = list(map(CollectedRule.from_json, data["rules"]))
        data.pop("rules", None)
        return cls(scenarios=scenarios, rules=rules, **data)


class CollectingFormatter(Formatter):
    """Formatter for collected scenarios, rules and features."""

    def __init__(self, stream_opener, config):
        super().__init__(stream_opener, config)

        self.current_feature: CollectedFeature | None = None
        self.current_rule: CollectedRule | None = None
        self.current_scenario: CollectedScenario | None = None
        self.current_step: CollectedStep | None = None

        self.steps_to_process: list[CollectedStep] = []
        self.current_step_text: list[str] = []

    def write_function(self):
        def formatter_write_line(*values, sep=""):
            line = sep.join(map(str, chain.from_iterable(values)))
            print(line)
            self.current_step_text.append(line)

        return formatter_write_line

    def feature(self, feature: Feature):
        self.current_feature = CollectedFeature(feature.filename, feature.name, feature.description, feature.tags)

    def rule(self, rule):
        self.current_rule = CollectedRule(rule.name, rule.description, rule.tags)
        self.current_feature.rules.append(self.current_rule)

    def background(self, background):
        pass

    def scenario(self, scenario: Scenario):
        self.current_scenario = CollectedScenario(scenario.name, scenario.tags)
        if not self.current_rule:
            assert self.current_feature
            self.current_feature.scenarios.append(self.current_scenario)
        else:
            self.current_rule.scenarios.append(self.current_scenario)
        self.steps_to_process = []

    def step(self, step: Step):
        """Append step to scenario."""
        step_to_store = CollectedStep(step.name, step.keyword)
        assert self.current_scenario
        self.current_scenario.steps.append(step_to_store)

        self.steps_to_process.append(step_to_store)

    def match(self, match):
        """Match next step."""
        self.current_step = self.steps_to_process.pop(0)
        self.current_step_text = []

    def result(self, step_result):
        """Result."""
        self.current_step.status = step_result.status.name
        self.current_step.text = self.current_step_text
        self.current_step.error_message = step_result.error_message

        self.current_scenario.status = step_result.status.name

    def eof(self):
        """End of file reached."""
        self.current_feature.finished()



"""Template module using jinja2."""

import pathlib

from jinja2 import Environment
from jinja2 import FileSystemLoader


current_directory = pathlib.Path(__file__).parent.resolve()
env = Environment(autoescape=True, loader=FileSystemLoader(current_directory, followlinks=True))


def convert_template(filename, context):
    """Convert json context to output format with jinja."""
    template = env.get_template(filename)
    return template.render(**context)



"""Module for generating user needs page."""

import sys

from .template import convert_template
from .userneeds import load_userneeds_from_file


def main(args=None):
    """Load input csv file with user needs."""
    if args is None:
        args = sys.argv

    if len(args) != 3:
        print(f"Invalid command format, format is:\n {args[0]} <userneed text file> <userneeds sphinx file>\n")
        sys.exit(1)

    userneeds = load_userneeds_from_file(args[1])

    context = {"userneeds": userneeds}

    with open(args[2], "w") as file:
        file.write(convert_template("userneeds.jinja2", context))


if __name__ == "__main__":
    main()



"""cli entry point for livingdocs.

Parse command line arguments, invoke proper methods.
"""

import argparse
import logging
import sys

from . import __version__


def main(args=None):
    """Entry point for aidee-living-docs cli.

    The .toml entry_point wraps this in sys.exit already so this effectively
    becomes sys.exit(main()).
    The __main__ entry point similarly wraps sys.exit().
    """
    if args is None:
        args = sys.argv[1:]

    parsed_args = __get_args(args)
    logging.basicConfig(
        level=getattr(logging, parsed_args.log_level.upper(), logging.INFO),
        format="%(asctime)s:%(levelname)s:%(message)s",
    )


def __get_args(args):
    """Parse arguments passed in from shell."""
    return __get_parser().parse_args(args)


def __get_parser():
    """Return ArgumentParser."""
    parser = argparse.ArgumentParser(
        allow_abbrev=True,
        description="Aidee Living Documentation application",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    log_levels = ["CRITICAL", "WARNING", "INFO", "DEBUG"]
    parser.add_argument(
        "--log-level",
        help=f"Log level ({log_levels})",
        choices=log_levels,
        default="WARNING",
    )
    parser.add_argument(
        "--version",
        action="version",
        help="Echo version number.",
        version=f"{__version__}",
    )
    return parser


if __name__ == "__main__":
    main()



"""Default execution entry point if running the package via python -m."""

import sys

from . import cli


def main():
    """Run aidee-living-docs from script entry point."""
    return cli.main()


if __name__ == "__main__":
    sys.exit(main())



"""Create traceability. Link scenarios with status to user needs."""

import sys

from .collecting_formatter import CollectedScenario
from .feature_file import load_all_feature_files_in_directory
from .feature_file import status_to_style
from .template import convert_template
from .userneeds import load_userneeds_from_file


def add_scenario_to_userneed(userneeds, user_need_id, scenario_id, scenario_name, result):
    """Add scenario to proper user need."""
    user_need_id = user_need_id.lower()
    for userneed in userneeds:
        if userneed["id"].lower() == user_need_id:
            userneed["scenarios"].append({"id": scenario_id, "name": scenario_name, "result": result})


def userneed_id_for_scenario(scenario: CollectedScenario):
    """Find user need (un_-prefixed) in scenario."""
    for tag in scenario.tags:
        if tag.lower().startswith("un_"):
            return tag.lower()

    return ""


def id_for_scenario(scenario: CollectedScenario):
    """Find id tag for scenario (starting with id_)."""
    for tag in scenario.tags:
        if tag.lower().startswith("id_"):
            return tag

    return ""


def main(args=None):
    """Script entry point."""
    if args is None:
        args = sys.argv

    if len(args) != 4:
        print(
            f"Invalid command format, format is:\n {args[0]} <userneed text file> "
            "<path to feature file results> <output file>\n"
        )
        sys.exit(1)

    userneeds = load_userneeds_from_file(args[1])
    for userneed in userneeds:
        userneed["scenarios"] = []

    features = load_all_feature_files_in_directory(args[2])
    for feature in features:
        for scenario in feature.scenarios:
            user_need_id = userneed_id_for_scenario(scenario)
            scenario_id = id_for_scenario(scenario)
            result = scenario.status
            if user_need_id != "" and scenario_id != "":
                add_scenario_to_userneed(
                    userneeds,
                    user_need_id,
                    scenario_id,
                    feature.name + " - " + scenario.name,
                    result,
                )

    context = {"userneeds": userneeds, "status_to_style": status_to_style}

    with open(args[3], "w") as file:
        file.write(convert_template("userneeds_trace.jinja2", context))


if __name__ == "__main__":
    """Relay to main method when invoked as standalone script."""
    main()



"""Initialize the livingdocs package."""

import importlib.metadata

from .json_formatter import JSONFormatter


__all__ = [
    "JSONFormatter",
]

try:
    __version__ = importlib.metadata.version("aidee-living-docs")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"
